/*
 * PtokaX - hub server for Direct Connect peer to peer network.

 * Copyright (C) 2002-2005  Ptaczek, Ptaczek at PtokaX dot org
 * Copyright (C) 2004-2014  Petr Kozelka, PPK at PtokaX dot org

 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 3
 * as published by the Free Software Foundation.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

//---------------------------------------------------------------------------
#include "stdinc.h"
//---------------------------------------------------------------------------
#include "hashRegManager.h"
//---------------------------------------------------------------------------
#include "colUsers.h"
#include "GlobalDataQueue.h"
#include "hashUsrManager.h"
#include "LanguageManager.h"
#include "ProfileManager.h"
#include "ServerManager.h"
#include "SettingManager.h"
#include "User.h"
#include "utility.h"
//---------------------------------------------------------------------------
#ifdef _WIN32
	#pragma hdrstop
#endif
//---------------------------------------------------------------------------
#include "../core/PXBReader.h"
//---------------------------------------------------------------------------
#ifdef _BUILD_GUI
    #include "../gui.win/RegisteredUserDialog.h"
    #include "../gui.win/RegisteredUsersDialog.h"
#endif
//---------------------------------------------------------------------------
#ifdef HASH_PASS
	#include "../sha2/sha2.h"
#endif
//---------------------------------------------------------------------------
clsRegManager * clsRegManager::mPtr = NULL;
//---------------------------------------------------------------------------

// alex82 ...     
RegUser::RegUser(char * Nick, char * Pass, const uint16_t &iRegProfile, const time_t &RegDate, const time_t &LastEnter, const time_t &OnlineTime) {
    prev = NULL;
    next = NULL;
    
    size_t szNickLen = strlen(Nick);
#ifdef _WIN32
    sNick = (char *)HeapAlloc(clsServerManager::hPtokaXHeap, HEAP_NO_SERIALIZE, szNickLen+1);
#else
	sNick = (char *)malloc(szNickLen+1);
#endif
    if(sNick == NULL) {
        AppendDebugLog("%s - [MEM] Cannot allocate %" PRIu64 " bytes for sNick in RegUser::RegUser\n", (uint64_t)(szNickLen+1));

        return;
    }   
    memcpy(sNick, Nick, szNickLen);
    sNick[szNickLen] = '\0';
    
    size_t szPassLen = strlen(Pass);
#ifdef _WIN32
    sPass = (char *)HeapAlloc(clsServerManager::hPtokaXHeap, HEAP_NO_SERIALIZE, szPassLen+1);
#else
	sPass = (char *)malloc(szPassLen+1);
#endif
    if(sPass == NULL) {
		AppendDebugLog("%s - [MEM] Cannot allocate %" PRIu64 " bytes for sPass in RegUser::RegUser\n", (uint64_t)(szPassLen+1));

        return;
    }   
    memcpy(sPass, Pass, szPassLen);
    sPass[szPassLen] = '\0';

    tLastBadPass = 0;
    ui8BadPassCount = 0;
    
    ui16Profile = iRegProfile;
	// alex82 ...     
	tRegDate = RegDate;
	tLastEnter = LastEnter;
	tOnlineTime = OnlineTime;
	sCustom = NULL;
	ui32Hash = HashNick(sNick, szNickLen);
    hashtableprev = NULL;
    hashtablenext = NULL;
}
//---------------------------------------------------------------------------

RegUser::~RegUser(void) {
#ifdef _WIN32
    if(sNick != NULL && HeapFree(clsServerManager::hPtokaXHeap, HEAP_NO_SERIALIZE, (void *)sNick) == 0) {
		AppendDebugLog("%s - [MEM] Cannot deallocate sNick in RegUser::~RegUser\n", 0);
    }
#else
	free(sNick);
#endif
    sNick = NULL;

#ifdef _WIN32
    if(sPass != NULL && HeapFree(clsServerManager::hPtokaXHeap, HEAP_NO_SERIALIZE, (void *)sPass) == 0) {
		AppendDebugLog("%s - [MEM] Cannot deallocate sPass in RegUser::~RegUser\n", 0);
    }
#else
	free(sPass);
#endif
    sPass = NULL;
	
#ifdef _WIN32
	if(sCustom != NULL && HeapFree(clsServerManager::hPtokaXHeap, HEAP_NO_SERIALIZE, (void *)sCustom) == 0) {
		AppendDebugLog("%s - [MEM] Cannot deallocate sCustom in RegUser::~RegUser\n", 0);
	}
#else
	free(sCustom);
#endif
	sCustom = NULL;

}
//--------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------

clsRegManager::clsRegManager(void) {
    RegListS = RegListE = NULL;

    for(uint32_t ui32i = 0; ui32i < 65536; ui32i++) {
        table[ui32i] = NULL;
    }

    ui8SaveCalls = 0;
}
//---------------------------------------------------------------------------

clsRegManager::~clsRegManager(void) {
    RegUser *next = RegListS;
        
    while(next != NULL) {
        RegUser *curReg = next;
		next = curReg->next;
		delete curReg;
    }
}
//---------------------------------------------------------------------------

bool clsRegManager::AddNew(char * sNick, char * sPasswd, const uint16_t &iProfile) {
    if(Find(sNick, strlen(sNick)) != NULL) {
        return false;
    }

	// alex82 ...     
	time_t curtime = time(NULL);

#ifdef HASH_PASS
	char buff[SHA256_DIGEST_BASE64_LENGTH];
	SHA256_Base64((unsigned char *)sPasswd, strlen(sPasswd), (unsigned char *)buff);
    RegUser * pNewUser = new RegUser(sNick, buff, iProfile, curtime, curtime, 0);
#else
    RegUser * pNewUser = new RegUser(sNick, sPasswd, iProfile, curtime, curtime, 0);
#endif

    if(pNewUser == NULL || pNewUser->sNick == NULL || pNewUser->sPass == NULL) {
        if(pNewUser == NULL) {
            AppendDebugLog("%s - [MEM] Cannot allocate pNewUser in hashRegMan::AddNew\n", 0);
        } else if(pNewUser->sNick == NULL) {
            delete pNewUser;
            AppendDebugLog("%s - [MEM] Cannot allocate pNewUser->sNick in hashRegMan::AddNew\n", 0);
        } else if(pNewUser->sPass) {
            delete pNewUser;
            AppendDebugLog("%s - [MEM] Cannot allocate pNewUser->sPass in hashRegMan::AddNew\n", 0);
        }

        return false;
    }

	Add(pNewUser);

#ifdef _BUILD_GUI
    if(clsRegisteredUsersDialog::mPtr != NULL) {
        clsRegisteredUsersDialog::mPtr->AddReg(pNewUser);
    }
#endif

    Save(true);

    if(clsServerManager::bServerRunning == false) {
        return true;
    }

	User * AddedUser = clsHashManager::mPtr->FindUser(pNewUser->sNick, strlen(pNewUser->sNick));

    if(AddedUser != NULL) {
        bool bAllowedOpChat = clsProfileManager::mPtr->IsAllowed(AddedUser, clsProfileManager::ALLOWEDOPCHAT);
        AddedUser->iProfile = iProfile;

        if(((AddedUser->ui32BoolBits & User::BIT_OPERATOR) == User::BIT_OPERATOR) == false) {
            if(clsProfileManager::mPtr->IsAllowed(AddedUser, clsProfileManager::HASKEYICON) == true) {
                AddedUser->ui32BoolBits |= User::BIT_OPERATOR;
            } else {
                AddedUser->ui32BoolBits &= ~User::BIT_OPERATOR;
            }

            if(((AddedUser->ui32BoolBits & User::BIT_OPERATOR) == User::BIT_OPERATOR) == true) {
				// alex82 ...   
				if(((AddedUser->ui32InfoBits & User::INFOBIT_HIDE_KEY) == User::INFOBIT_HIDE_KEY) == false) {
					clsUsers::mPtr->Add2OpList(AddedUser);
					clsGlobalDataQueue::mPtr->OpListStore(AddedUser->sNick);
				}

                if(bAllowedOpChat != clsProfileManager::mPtr->IsAllowed(AddedUser, clsProfileManager::ALLOWEDOPCHAT)) {
					if(clsSettingManager::mPtr->bBools[SETBOOL_REG_OP_CHAT] == true &&
                        (clsSettingManager::mPtr->bBools[SETBOOL_REG_BOT] == false || clsSettingManager::mPtr->bBotsSameNick == false)) {
                        if(((AddedUser->ui32SupportBits & User::SUPPORTBIT_NOHELLO) == User::SUPPORTBIT_NOHELLO) == false) {
                            AddedUser->SendCharDelayed(clsSettingManager::mPtr->sPreTexts[clsSettingManager::SETPRETXT_OP_CHAT_HELLO],
                                clsSettingManager::mPtr->ui16PreTextsLens[clsSettingManager::SETPRETXT_OP_CHAT_HELLO]);
                        }

                        AddedUser->SendCharDelayed(clsSettingManager::mPtr->sPreTexts[clsSettingManager::SETPRETXT_OP_CHAT_MYINFO],
                            clsSettingManager::mPtr->ui16PreTextsLens[clsSettingManager::SETPRETXT_OP_CHAT_MYINFO]);

						char msg[128];
						int imsgLen = sprintf(msg, "$OpList %s$$|", clsSettingManager::mPtr->sTexts[SETTXT_OP_CHAT_NICK]);
                        if(CheckSprintf(imsgLen, 128, "clsRegManager::AddNew") == true) {
                            AddedUser->SendCharDelayed(msg, imsgLen);
                        }
                    }
                }
            }
        }
    }

    return true;
}
//---------------------------------------------------------------------------

void clsRegManager::Add(RegUser * Reg) {
	Add2Table(Reg);
    
    if(RegListE == NULL) {
    	RegListS = Reg;
    	RegListE = Reg;
    } else {
        Reg->prev = RegListE;
    	RegListE->next = Reg;
        RegListE = Reg;
    }

	return;
}
//---------------------------------------------------------------------------

void clsRegManager::Add2Table(RegUser * Reg) {
    uint16_t ui16dx = 0;
    memcpy(&ui16dx, &Reg->ui32Hash, sizeof(uint16_t));

    if(table[ui16dx] != NULL) {
        table[ui16dx]->hashtableprev = Reg;
        Reg->hashtablenext = table[ui16dx];
    }
    
    table[ui16dx] = Reg;
}
//---------------------------------------------------------------------------

void clsRegManager::ChangeReg(RegUser * pReg, char * sNewPasswd, const uint16_t &ui16NewProfile) {
	// alex82 ...    
    if(sNewPasswd != NULL && strcmp(pReg->sPass, sNewPasswd) != 0) {
        size_t szPassLen = strlen(sNewPasswd);

#ifdef HASH_PASS
		SHA256_Base64((unsigned char *)sNewPasswd, szPassLen, (unsigned char *)pReg->sPass);
#else
        char * sOldPass = pReg->sPass;
#ifdef _WIN32
        pReg->sPass = (char *)HeapReAlloc(clsServerManager::hPtokaXHeap, HEAP_NO_SERIALIZE, (void *)sOldPass, szPassLen+1);
#else
		pReg->sPass = (char *)realloc(sOldPass, szPassLen+1);
#endif
        if(pReg->sPass == NULL) {
            pReg->sPass = sOldPass;

			AppendDebugLog("%s - [MEM] Cannot reallocate %" PRIu64 " bytes for sPass in hashRegMan::ChangeReg\n", (uint64_t)(szPassLen+1));

            return;
        }
        memcpy(pReg->sPass, sNewPasswd, szPassLen);
        pReg->sPass[szPassLen] = '\0';
#endif
    }

    pReg->ui16Profile = ui16NewProfile;

#ifdef _BUILD_GUI
    if(clsRegisteredUsersDialog::mPtr != NULL) {
        clsRegisteredUsersDialog::mPtr->RemoveReg(pReg);
        clsRegisteredUsersDialog::mPtr->AddReg(pReg);
    }
#endif

    clsRegManager::mPtr->Save(true);

    if(clsServerManager::bServerRunning == false) {
        return;
    }

    User *ChangedUser = clsHashManager::mPtr->FindUser(pReg->sNick, strlen(pReg->sNick));
    if(ChangedUser != NULL && ChangedUser->iProfile != (int32_t)ui16NewProfile) {
        char msg[128];
        bool bAllowedOpChat = clsProfileManager::mPtr->IsAllowed(ChangedUser, clsProfileManager::ALLOWEDOPCHAT);

        ChangedUser->iProfile = (int32_t)ui16NewProfile;

        if(((ChangedUser->ui32BoolBits & User::BIT_OPERATOR) == User::BIT_OPERATOR) !=
            clsProfileManager::mPtr->IsAllowed(ChangedUser, clsProfileManager::HASKEYICON)) {
            if(clsProfileManager::mPtr->IsAllowed(ChangedUser, clsProfileManager::HASKEYICON) == true) {
                ChangedUser->ui32BoolBits |= User::BIT_OPERATOR;
				// alex82 ...   
				if(((ChangedUser->ui32InfoBits & User::INFOBIT_HIDE_KEY) == User::INFOBIT_HIDE_KEY) == false) {
	                clsUsers::mPtr->Add2OpList(ChangedUser);
					clsGlobalDataQueue::mPtr->OpListStore(ChangedUser->sNick);
				}
            } else {
                ChangedUser->ui32BoolBits &= ~User::BIT_OPERATOR;
				// alex82 ...   
				if(((ChangedUser->ui32InfoBits & User::INFOBIT_HIDE_KEY) == User::INFOBIT_HIDE_KEY) == false) {
 					// alex82 ...   OpList
                   int imsgLen = sprintf(msg, "$Quit %s|", ChangedUser->sNick);
                    if(CheckSprintf(imsgLen, 128, "clsRegManager::ChangeReg1") == true) {
						clsGlobalDataQueue::mPtr->AddQueueItem(msg, imsgLen, NULL, 0, clsGlobalDataQueue::CMD_QUIT);
                    }
					switch(clsSettingManager::mPtr->ui8FullMyINFOOption) {
						case 0:
							clsGlobalDataQueue::mPtr->AddQueueItem(ChangedUser->sMyInfoLong, ChangedUser->ui16MyInfoLongLen, NULL, 0, clsGlobalDataQueue::CMD_MYINFO);
							break;
						case 1:
							clsGlobalDataQueue::mPtr->AddQueueItem(ChangedUser->sMyInfoShort, ChangedUser->ui16MyInfoShortLen, ChangedUser->sMyInfoLong, ChangedUser->ui16MyInfoLongLen, clsGlobalDataQueue::CMD_MYINFO);
							break;
						case 2:
							clsGlobalDataQueue::mPtr->AddQueueItem(ChangedUser->sMyInfoShort, ChangedUser->ui16MyInfoShortLen, NULL, 0, clsGlobalDataQueue::CMD_MYINFO);
							break;
						default:
							break;
					}
					clsUsers::mPtr->DelFromOpList(ChangedUser->sNick);
				}
            }
        }

        if(bAllowedOpChat != clsProfileManager::mPtr->IsAllowed(ChangedUser, clsProfileManager::ALLOWEDOPCHAT)) {
            if(clsProfileManager::mPtr->IsAllowed(ChangedUser, clsProfileManager::ALLOWEDOPCHAT) == true) {
                if(clsSettingManager::mPtr->bBools[SETBOOL_REG_OP_CHAT] == true &&
                    (clsSettingManager::mPtr->bBools[SETBOOL_REG_BOT] == false || clsSettingManager::mPtr->bBotsSameNick == false)) {
                    if(((ChangedUser->ui32SupportBits & User::SUPPORTBIT_NOHELLO) == User::SUPPORTBIT_NOHELLO) == false) {
                        ChangedUser->SendCharDelayed(clsSettingManager::mPtr->sPreTexts[clsSettingManager::SETPRETXT_OP_CHAT_HELLO],
                        clsSettingManager::mPtr->ui16PreTextsLens[clsSettingManager::SETPRETXT_OP_CHAT_HELLO]);
                    }

                    ChangedUser->SendCharDelayed(clsSettingManager::mPtr->sPreTexts[clsSettingManager::SETPRETXT_OP_CHAT_MYINFO],
                        clsSettingManager::mPtr->ui16PreTextsLens[clsSettingManager::SETPRETXT_OP_CHAT_MYINFO]);

                    int imsgLen = sprintf(msg, "$OpList %s$$|", clsSettingManager::mPtr->sTexts[SETTXT_OP_CHAT_NICK]);
                    if(CheckSprintf(imsgLen, 128, "clsRegManager::ChangeReg1") == true) {
                        ChangedUser->SendCharDelayed(msg, imsgLen);
                    }
                }
            } else {
                if(clsSettingManager::mPtr->bBools[SETBOOL_REG_OP_CHAT] == true && (clsSettingManager::mPtr->bBools[SETBOOL_REG_BOT] == false || clsSettingManager::mPtr->bBotsSameNick == false)) {
                    int imsgLen = sprintf(msg, "$Quit %s|", clsSettingManager::mPtr->sTexts[SETTXT_OP_CHAT_NICK]);
                    if(CheckSprintf(imsgLen, 128, "clsRegManager::ChangeReg2") == true) {
                        ChangedUser->SendCharDelayed(msg, imsgLen);
                    }
                }
            }
        }
    }

#ifdef _BUILD_GUI
    if(clsRegisteredUserDialog::mPtr != NULL) {
        clsRegisteredUserDialog::mPtr->RegChanged(pReg);
    }
#endif
}
//---------------------------------------------------------------------------
// alex82 ...      
void clsRegManager::SetCustom(RegUser * pReg, char * sCustom, const size_t &szCustomLen) {
	if (szCustomLen == 0) {
		if (pReg->sCustom != NULL) {
#ifdef _WIN32
			if(HeapFree(clsServerManager::hPtokaXHeap, HEAP_NO_SERIALIZE, (void *)pReg->sCustom) == 0) {
				AppendDebugLog("%s - [MEM] Cannot deallocate sCustom in clsRegManager::SetCustom\n", 0);
		    }
#else
			free(pReg->sCustom);
#endif
			pReg->sCustom = NULL;
		}
	} else {
		if (pReg->sCustom != NULL) {
			if (szCustomLen != strlen(pReg->sCustom)) {
				char * sOldCustom = pReg->sCustom;
#ifdef _WIN32
				pReg->sCustom = (char *)HeapReAlloc(clsServerManager::hPtokaXHeap, HEAP_NO_SERIALIZE, (void *)sOldCustom, szCustomLen+1);
#else
				pReg->sCustom = (char *)realloc(sOldCustom, szCustomLen+1);
#endif
			}
		} else {
#ifdef _WIN32
			pReg->sCustom = (char *)HeapAlloc(clsServerManager::hPtokaXHeap, HEAP_NO_SERIALIZE, szCustomLen+1);
#else
			pReg->sCustom = (char *)malloc(szCustomLen+1);
#endif
		}
        memcpy(pReg->sCustom, sCustom, szCustomLen);
        pReg->sCustom[szCustomLen] = '\0';
	}
}

//---------------------------------------------------------------------------

#ifdef _BUILD_GUI
void clsRegManager::Delete(RegUser * pReg, const bool &bFromGui/* = false*/) {
#else
void clsRegManager::Delete(RegUser * pReg, const bool &/*bFromGui = false*/) {
#endif
	if(clsServerManager::bServerRunning == true) {
        User * pRemovedUser = clsHashManager::mPtr->FindUser(pReg->sNick, strlen(pReg->sNick));

        if(pRemovedUser != NULL) {
            pRemovedUser->iProfile = -1;
            if(((pRemovedUser->ui32BoolBits & User::BIT_OPERATOR) == User::BIT_OPERATOR) == true) {
                clsUsers::mPtr->DelFromOpList(pRemovedUser->sNick);
                pRemovedUser->ui32BoolBits &= ~User::BIT_OPERATOR;

                if(clsSettingManager::mPtr->bBools[SETBOOL_REG_OP_CHAT] == true && (clsSettingManager::mPtr->bBools[SETBOOL_REG_BOT] == false || clsSettingManager::mPtr->bBotsSameNick == false)) {
                    char msg[128];
                    int imsgLen = sprintf(msg, "$Quit %s|", clsSettingManager::mPtr->sTexts[SETTXT_OP_CHAT_NICK]);
                    if(CheckSprintf(imsgLen, 128, "clsRegManager::Delete") == true) {
                        pRemovedUser->SendCharDelayed(msg, imsgLen);
                    }
                }
            }
        }
    }

#ifdef _BUILD_GUI
    if(bFromGui == false && clsRegisteredUsersDialog::mPtr != NULL) {
        clsRegisteredUsersDialog::mPtr->RemoveReg(pReg);
    }
#endif

	Rem(pReg);

#ifdef _BUILD_GUI
    if(clsRegisteredUserDialog::mPtr != NULL) {
        clsRegisteredUserDialog::mPtr->RegDeleted(pReg);
    }
#endif

    delete pReg;

    Save(true);
}
//---------------------------------------------------------------------------

void clsRegManager::Rem(RegUser * Reg) {
	RemFromTable(Reg);
    
    RegUser *prev, *next;
    prev = Reg->prev; next = Reg->next;

    if(prev == NULL) {
        if(next == NULL) {
            RegListS = NULL;
            RegListE = NULL;
        } else {
            next->prev = NULL;
            RegListS = next;
        }
    } else if(next == NULL) {
        prev->next = NULL;
        RegListE = prev;
    } else {
        prev->next = next;
        next->prev = prev;
    }
}
//---------------------------------------------------------------------------

void clsRegManager::RemFromTable(RegUser * Reg) {
    if(Reg->hashtableprev == NULL) {
        uint16_t ui16dx = 0;
        memcpy(&ui16dx, &Reg->ui32Hash, sizeof(uint16_t));

        if(Reg->hashtablenext == NULL) {
            table[ui16dx] = NULL;
        } else {
            Reg->hashtablenext->hashtableprev = NULL;
			table[ui16dx] = Reg->hashtablenext;
        }
    } else if(Reg->hashtablenext == NULL) {
        Reg->hashtableprev->hashtablenext = NULL;
    } else {
        Reg->hashtableprev->hashtablenext = Reg->hashtablenext;
        Reg->hashtablenext->hashtableprev = Reg->hashtableprev;
    }

	Reg->hashtableprev = NULL;
    Reg->hashtablenext = NULL;
}
//---------------------------------------------------------------------------

RegUser* clsRegManager::Find(char * sNick, const size_t &szNickLen) {
    uint32_t ui32Hash = HashNick(sNick, szNickLen);

    uint16_t ui16dx = 0;
    memcpy(&ui16dx, &ui32Hash, sizeof(uint16_t));

    RegUser *next = table[ui16dx];

    while(next != NULL) {
        RegUser *cur = next;
        next = cur->hashtablenext;

		if(cur->ui32Hash == ui32Hash && strcasecmp(cur->sNick, sNick) == 0) {
            return cur;
        }
    }

    return NULL;
}
//---------------------------------------------------------------------------

RegUser* clsRegManager::Find(User * u) {
    uint16_t ui16dx = 0;
    memcpy(&ui16dx, &u->ui32NickHash, sizeof(uint16_t));

	RegUser *next = table[ui16dx];

    while(next != NULL) {
        RegUser *cur = next;
        next = cur->hashtablenext;

		if(cur->ui32Hash == u->ui32NickHash && strcasecmp(cur->sNick, u->sNick) == 0) {
            return cur;
        }
    }

    return NULL;
}
//---------------------------------------------------------------------------

RegUser* clsRegManager::Find(uint32_t ui32Hash, char * sNick) {
    uint16_t ui16dx = 0;
    memcpy(&ui16dx, &ui32Hash, sizeof(uint16_t));

	RegUser *next = table[ui16dx];

    while(next != NULL) {
        RegUser *cur = next;
        next = cur->hashtablenext;

		if(cur->ui32Hash == ui32Hash && strcasecmp(cur->sNick, sNick) == 0) {
            return cur;
        }
    }

    return NULL;
}
//---------------------------------------------------------------------------

// alex82 ...          
void clsRegManager::UpdateTimes(void) {
	User *next = clsUsers::mPtr->llist;
	time_t curtime = time(NULL);

	while(next != NULL) {
		User *curUser = next;
		next = curUser->next;
		if(curUser->ui8State == User::STATE_ADDED && curUser->iProfile != -1) {
			RegUser * pReg = clsRegManager::mPtr->Find(curUser);
			if(pReg != NULL) {
				// alex82 ...   ,      ...
				if (curtime >= pReg->tLastEnter) {
					pReg->tOnlineTime = pReg->tOnlineTime+curtime-pReg->tLastEnter;
				}
				pReg->tLastEnter = curtime;
			}
		}
	}
}
//---------------------------------------------------------------------------

void clsRegManager::Load(void) {
	// alex82 ...       XML
    uint16_t iProfilesCount = (uint16_t)(clsProfileManager::mPtr->iProfileCount-1);
    bool bIsBuggy = false;
	time_t curtime = time(NULL);

#ifdef _WIN32
    TiXmlDocument doc((clsServerManager::sPath+"\\cfg\\RegisteredUsers.xml").c_str());
#else
	TiXmlDocument doc((clsServerManager::sPath+"/cfg/RegisteredUsers.xml").c_str());
#endif

    if(doc.LoadFile()) {
        TiXmlHandle cfg(&doc);
        TiXmlNode *registeredusers = cfg.FirstChild("RegisteredUsers").Node();
        if(registeredusers != NULL) {
            TiXmlNode *child = NULL;
            while((child = registeredusers->IterateChildren(child)) != NULL) {
				TiXmlNode *registereduser = child->FirstChild("Nick");

				if(registereduser == NULL || (registereduser = registereduser->FirstChild()) == NULL) {
					continue;
				}

                char *nick = (char *)registereduser->Value();
                
				if(strlen(nick) > 64 || (registereduser = child->FirstChild("Password")) == NULL ||
                    (registereduser = registereduser->FirstChild()) == NULL) {
					continue;
				}

                char *pass = (char *)registereduser->Value();
                
				if(strlen(pass) > 64 || (registereduser = child->FirstChild("Profile")) == NULL ||
                    (registereduser = registereduser->FirstChild()) == NULL) {
					continue;
				}

				uint16_t iProfile = (uint16_t)atoi(registereduser->Value());

				// alex82 ...     
				time_t regdate,lastenter,onlinetime;
				if ((registereduser = child->FirstChild("RegDate")) != NULL && (registereduser = registereduser->FirstChild()) != NULL)
					regdate = (time_t)atol(registereduser->Value());
				else {
					regdate = curtime;
					bIsBuggy = true;
				}
				if ((registereduser = child->FirstChild("LastEnter")) != NULL && (registereduser = registereduser->FirstChild()) != NULL)
					lastenter = (time_t)atol(registereduser->Value());
				else {
					lastenter = curtime;
					bIsBuggy = true;
				}
				if ((registereduser = child->FirstChild("OnlineTime")) != NULL && (registereduser = registereduser->FirstChild()) != NULL)
					onlinetime = (time_t)atol(registereduser->Value());
				else {
					onlinetime = 0;
					bIsBuggy = true;
				}

				if(iProfile > iProfilesCount) {
                    char msg[1024];
                    int imsgLen = sprintf(msg, "%s %s %s! %s %s.", clsLanguageManager::mPtr->sTexts[LAN_USER], nick, clsLanguageManager::mPtr->sTexts[LAN_HAVE_NOT_EXIST_PROFILE],
                        clsLanguageManager::mPtr->sTexts[LAN_CHANGED_PROFILE_TO], clsProfileManager::mPtr->ProfilesTable[iProfilesCount]->sName);
					CheckSprintf(imsgLen, 1024, "clsRegManager::Load");

#ifdef _BUILD_GUI
					::MessageBox(NULL, msg, clsServerManager::sTitle.c_str(), MB_OK | MB_ICONEXCLAMATION);
#else
					AppendLog(msg);
#endif

                    iProfile = iProfilesCount;
                    bIsBuggy = true;
                }

                if(Find((char*)nick, strlen(nick)) == NULL) {
#ifdef HASH_PASS
					if (pass[0] != '|' || strlen(pass) != SHA256_DIGEST_BASE64_LENGTH)
					{
						char buff[SHA256_DIGEST_BASE64_LENGTH];
						SHA256_Base64((unsigned char *)pass, strlen(pass), (unsigned char *)buff);
						pass = buff;
						bIsBuggy = true;
					}
					else pass = pass+1;
#endif
					// alex82 ...     
                    RegUser *newUser = new RegUser(nick, pass, iProfile, regdate, lastenter, onlinetime);
                    if(newUser == NULL) {
                    	string sDbgstr = "%s - [BUF] Cannot allocate newUser in hashRegMan::LoadXmlRegList!";
#ifdef _WIN32
						sDbgstr += " "+string(HeapValidate(GetProcessHeap, 0, 0))+GetMemStat()+"\n";
#else
						sDbgstr += "\n";
#endif
						AppendDebugLog(sDbgstr.c_str(),0);
                    	exit(EXIT_FAILURE);
                    }
					if ((registereduser = child->FirstChild("Custom")) != NULL && (registereduser = registereduser->FirstChild()) != NULL) {
						char *custom = (char *)registereduser->Value();
						size_t custom_len = strlen(custom);
						if (custom_len > 0 && custom_len <= 256)
							SetCustom(newUser, custom, custom_len);
					}
					Add(newUser);
                } else {
                    char msg[1024];
                    int imsgLen = sprintf(msg, "%s %s %s! %s.", clsLanguageManager::mPtr->sTexts[LAN_USER], nick, clsLanguageManager::mPtr->sTexts[LAN_IS_ALREADY_IN_REGS], 
                        clsLanguageManager::mPtr->sTexts[LAN_USER_DELETED]);
					CheckSprintf(imsgLen, 1024, "hashRegMan::LoadXmlRegList2");

#ifdef _BUILD_GUI
					::MessageBox(NULL, msg, clsServerManager::sTitle.c_str(), MB_OK | MB_ICONEXCLAMATION);
#else
					AppendLog(msg);
#endif

                    bIsBuggy = true;
                }
            }
			if(bIsBuggy == true)
				Save();
        }
    }
	// alex82 ... Error processing
	else {
		if (doc.Error() && doc.ErrorId() != TiXmlBase::TIXML_ERROR_OPENING_FILE){
			char msg[1024];
			int imsgLen = sprintf(msg, "Syntax error in cfg/RegisteredUsers.xml: %s Line: %d", doc.ErrorDesc(), doc.ErrorRow());
			CheckSprintf(imsgLen, 1024, "clsRegManager::Load");

#ifdef _BUILD_GUI
			::MessageBox(NULL, msg, clsServerManager::sTitle.c_str(), MB_OK | MB_ICONERROR);
#else
			AppendLog(msg);
#endif
			exit(EXIT_FAILURE);
		} else {
			Save();
		}
	}
}
//---------------------------------------------------------------------------

void clsRegManager::Save(const bool &bSaveOnChange/* = false*/, const bool &bSaveOnTime/* = false*/) {
    if(bSaveOnTime == true && ui8SaveCalls == 0) {
        return;
    }

    ui8SaveCalls++;

    if(bSaveOnChange == true && ui8SaveCalls < 100) {
        return;
    }

    ui8SaveCalls = 0;

// alex82 ...       XML
#ifdef HASH_PASS
	char buff[SHA256_DIGEST_BASE64_LENGTH+1];
	buff[0] = '|';
	buff[SHA256_DIGEST_BASE64_LENGTH] = '\0';
#endif

#ifdef _WIN32
    TiXmlDocument doc((clsServerManager::sPath+"\\cfg\\RegisteredUsers.xml").c_str());
#else
	TiXmlDocument doc((clsServerManager::sPath+"/cfg/RegisteredUsers.xml").c_str());
#endif
	// alex82 ... Charset changed
    doc.InsertEndChild(TiXmlDeclaration("1.0", "windows-1251", "yes"));
    TiXmlElement registeredusers("RegisteredUsers");
    RegUser *next = RegListS;
    while(next != NULL) {
        RegUser *curReg = next;
		next = curReg->next;
		
        TiXmlElement nick("Nick");
        nick.InsertEndChild(TiXmlText(curReg->sNick));
        
        TiXmlElement pass("Password");
#ifdef HASH_PASS
		memcpy(buff+1, curReg->sPass, SHA256_DIGEST_BASE64_LENGTH-1);
        pass.InsertEndChild(TiXmlText(buff));
#else
        pass.InsertEndChild(TiXmlText(curReg->sPass));
#endif
        
        TiXmlElement profile("Profile");
        profile.InsertEndChild(TiXmlText(string(curReg->ui16Profile).c_str()));

		// alex82 ...     
        TiXmlElement regdate("RegDate");
        regdate.InsertEndChild(TiXmlText(string(curReg->tRegDate).c_str()));

        TiXmlElement lastenter("LastEnter");
        lastenter.InsertEndChild(TiXmlText(string(curReg->tLastEnter).c_str()));

        TiXmlElement onlinetime("OnlineTime");
        onlinetime.InsertEndChild(TiXmlText(string(curReg->tOnlineTime).c_str()));

	    TiXmlElement custom("Custom");
		if (curReg->sCustom != NULL) {
	        custom.InsertEndChild(TiXmlText(curReg->sCustom));
		}
		
        TiXmlElement registereduser("RegisteredUser");
        registereduser.InsertEndChild(nick);
        registereduser.InsertEndChild(pass);
        registereduser.InsertEndChild(profile);
		// alex82 ...     
        registereduser.InsertEndChild(regdate);
        registereduser.InsertEndChild(lastenter);
		registereduser.InsertEndChild(onlinetime);
		// alex82 ...      
		if (curReg->sCustom != NULL) {
			registereduser.InsertEndChild(custom);
		}
        
        registeredusers.InsertEndChild(registereduser);
    }
    doc.InsertEndChild(registeredusers);
    doc.SaveFile();
}
//---------------------------------------------------------------------------
